//Baseado em
//https://www-igm.univ-mlv.fr/~lecroq/string/node6.html

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "Util.h"
#include "ShiftOr.h"

//Create the bit table and return the value to compare to in order to check if the pattern is in the text
static int createBitTable(char *pattern, int patternSize, unsigned int BitTable[]) {
    //correctMask is the maximum number that is right +1. Any number lower than this is right
    unsigned int actualMask, correctMask;
    int patternIndex;

    //Initialize table
    //Starts every letter with the mask '11...11'
    for (int i = 0; i < ALPHABET_SIZE; ++i)
        BitTable[i] = ~0;

    //Starts with the correct Mask being '00...00', and the actual '00...01'
    //Begins in the first index of the pattern we want to search and stops when the pattern ends
    //Every loop updates the pattern and shifts the bits from the actual mask the left once
    for (correctMask = patternIndex = 0, actualMask = 1; patternIndex < patternSize; ++patternIndex, actualMask <<= 1)  {
        //the pattern for that character will be the complement of the mask and keep everything else that was in it
        //the actual mask only has a single bit with value '1', therefore the OR will only add the actual bit :)
        BitTable[pattern[patternIndex]] &= ~actualMask;
        //The minimum correct value updates adding the new index with the value '1' in it
        correctMask |= actualMask;
    }
    //Then we create a '0' in the most significant bit and complement it
    //This will make everything except the most significant go to zero :)
    correctMask = ~(correctMask>>1);
    #ifdef DEBUG
        print_bits(correctMask);
    #endif
    return(correctMask);
}

void searchWithShiftOr(char *pattern, int patternSize, char *text, int textSize) {
    unsigned int correctMask, state;
    unsigned int BitTable[ALPHABET_SIZE];
    int foundMatch = FALSE;
    int i;
    if (patternSize > WORD_SIZE)
        fprintf(stderr, "%s", "ShiftOR: Use pattern size <= word size");

    correctMask = createBitTable(pattern, patternSize, BitTable);

    for (i = 0, state = ~0; i < textSize; ++i) {
        //Shifts the state to create a '0' in the least significant bit
        //OR it with the mask for the current character in the bit table
        state = (state<<1) | BitTable[text[i]];
        //If the actual state in lesser that the mask of the minimum acceptable answer, we found the pattern
        if (state < correctMask)
        {
            foundMatch = TRUE;
            printf("Found at index %d\n",i - patternSize + 1);
        }
    }
    if(foundMatch == FALSE)
        printf("The pattern was not found :(\n");
    printf("\n");
}
